/*
 *  Copyright 2008 Mikhail Kryshen
 *
 *  This file is part of Charamega.
 *
 *  Charamega is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Charamega is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Charamega.  If not, see <http://www.gnu.org/licenses/>.
 */

package kryshen.charamega;

import javafx.animation.*;

import java.lang.Math;
import java.lang.System;
import java.lang.Character;

/**
 * Playing field model.
 */
class FieldModel {
    static attribute STATUS_STOPPED = 0;
    static attribute STATUS_RUNNING = 1;
    static attribute STATUS_LOSE = 2;
    static attribute STATUS_WIN = 3;

    attribute cardSize:Integer;    
    attribute cards:CardModel[];
    
    attribute opened1:CardModel;
    attribute opened2:CardModel;
    
    attribute pairs:Integer;
    attribute cleared:Integer;
    attribute flips:Integer;
    attribute mismatches = bind flips - cleared;

    attribute startTime:Integer;
    attribute time:Integer;
    attribute timeLimit = 300000; // 5 min
    attribute mismatchLimit = 100;

    attribute status = STATUS_STOPPED;
    
    attribute timer:Timeline = Timeline {
        repeatCount: Timeline.INDEFINITE
        keyFrames:
        KeyFrame {
            time: 250ms
            action: updateTime
        }
    }

    function checkPair(): Void {
        flips++;
        
        if (opened1.value == opened2.value) {
            cleared++;
            opened1.solved = true;
            opened2.solved = true;
        }

        opened1.opened = false;
        opened2.opened = false;        
        
        opened1 = null;
        opened2 = null;

        if (cleared == pairs) {
            stopStatus(STATUS_WIN);
        } else if (mismatches > mismatchLimit) {
            stopStatus(STATUS_LOSE);
        }
    }
    
    function updateTime():Void { 
        time = (System.currentTimeMillis() - startTime) as Integer;

        if (time >= timeLimit) {
            time = timeLimit;
            stopStatus(STATUS_LOSE);
        }    
    }

    function start(numChars:Integer, 
                    fieldWidth:Integer, 
                    fieldHeight:Integer): Void {

        var tcards:CardModel[];

        var numCards = numChars * 2;

        var aspect = (fieldWidth as Number) / fieldHeight;
        var cols:Integer = Math.round(Math.sqrt(numCards * aspect)) as Integer;
        var rows:Integer = Math.round(Math.sqrt(numCards / aspect)) as Integer;
        
        while (cols * rows < numCards) {
            if (cols > rows) {
                cols++;
            } else {
                rows++;
            }
        }
        
        var gridWidth = (fieldWidth as Number) / cols;
        var gridHeight = (fieldHeight as Number) / rows;
        
        cardSize = (Math.min(gridWidth, gridHeight) / 1.42).intValue();
        
        var offsetY = (gridHeight - cardSize) / 2;
        
        var acols:Integer[] = for (i in [1..rows]) { cols };
        var cells = cols * rows;
        
        while (cells > numCards) {
            var row = Math.floor(Math.random() * rows).intValue();
            if (acols[row] > 0) {
                acols[row]--;
                cells--;
            }            
        }

        for (j in [0..rows-1]) {
            for (i in [0..acols[j]-1]) {
                gridWidth = (fieldWidth as Number) / acols[j];
                var offsetX = (gridWidth - cardSize) / 2;
                
                insert CardModel {
                    x: i * gridWidth + offsetX + (offsetX / 4 - Math.random() * offsetX / 2)
                    y: j * gridHeight + offsetY + (offsetY / 4 - Math.random() * offsetY / 2)
                    angle: -45 + Math.random() * 90 //Math.random() * 360
                    field: this
                } into tcards;
            }
        }                                

        shuffleCards(numChars, tcards);

        cards = tcards;       
        
        pairs = numChars;
        cleared = 0;
        flips = 0;        
        
        opened1 = null;
        opened2 = null;

        status = STATUS_RUNNING;

        startTime = System.currentTimeMillis() as Integer;
        updateTime();
        timer.start();
    }
    
    /**
     * Generate new layout of cards.
     */
    private function shuffleCards(numChars:Integer, tcards:CardModel[]): Void {                
        var values = [
            [0x03B1..0x03C9],0x03A9,            // greek
            [0x20A0..0x20B5],                   // currency
            0x2190,0x2194,0x21CC,               // arrows
            0x221A,0x221E,0x222B,0x222E,        // math
            0x2318,0x23CE,0x23CF,0x23E3,
            0x25A8,0x25A9,[0x25C6..0x25D0],0x25D4,
            [0x2600..0x261A],[0x2620..0x2630],
            [0x2638..0x2672],[0x267A..0x2689],[0x2690..0x269C],
            0x26A0, 0x26A1,  
            [0x2706..0x2709],[0x270C..0x270E],0x2744
        ];
        
        var source:Integer[];
        
        for (i in [1..numChars]) {
            var k = Math.floor(Math.random() * values.size()) as Integer;
            insert values[k] into source;
            insert values[k] into source;
            delete values[k];
        }
        
        for (c in tcards) {
            var k = Math.floor(Math.random() * source.size()) as Integer;
            c.value = source[k];
            delete source[k];
        }
    }

    function stopStatus(stopStatus:Integer):Void {
        timer.stop();       
        status = stopStatus;

        for (card in cards) {
            card.opened = true;
        }
    }

    function stop():Void {
        stopStatus(STATUS_STOPPED);
    }
}
